# Copyright 2025 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ml_collections.config_dict import ConfigDict

RUN_CONFIG = ConfigDict(
    {
        "mapping": {
            "mapping_polymer": True,
            "mapping_ligand": True,
            "res_id_alignments": True,
            "enumerate_all_anchors": True,
        },
        "metric": {
            "calc_clashes": True,
            "calc_lddt": True,
            "calc_dockq": True,
            "calc_rmsd": True,
            "calc_pb_valid": True,
            "lddt": {
                "eps": 1e-6,
                "nucleotide_threshold": 30.0,
                "non_nucleotide_threshold": 15.0,
            },
            "clashes": {
                "vdw_scale_factor": 0.5,
            },
        },
    }
)
